"""
This a docstring for the module.
"""

import datetime as dt_module
from datetime import date, datetime, timedelta
from typing import Union

import pytz

from chronomeleon.models.mapping_config import MappingConfig

_berlin = pytz.timezone("Europe/Berlin")


def _convert_source_date_or_datetime_to_aware_datetime(
    source_value: Union[date, datetime], config: MappingConfig
) -> datetime:
    """
    returns a datetime object which is aware of the timezone (i.e. not naive) and is an exclusive end
    regardless of whether the source was configured as an inclusive or exclusive end.
    """
    source_value_datetime: datetime  # a non-naive datetime (exclusive, if end)
    if isinstance(source_value, datetime):
        source_value_datetime = source_value
        if config.is_end and config.source.is_inclusive_end:
            assert config.source.resolution is not None  # ensured by the consistency check
            source_value_datetime += config.source.resolution
    elif isinstance(source_value, date):
        if config.is_end and config.source.is_inclusive_end:
            source_value_datetime = datetime.combine(source_value + timedelta(days=1), datetime.min.time())
        else:
            source_value_datetime = datetime.combine(source_value, datetime.min.time())
    else:
        raise ValueError(f"source_value must be a date or datetime object but is {source_value.__class__.__name__}")
    if source_value_datetime.tzinfo is None:
        if config.source.implicit_timezone is not None:
            source_value_datetime = config.source.implicit_timezone.localize(source_value_datetime)
        else:
            # pylint:disable=line-too-long
            raise ValueError(
                "source_value must be timezone-aware or implicit_timezone must be set in the mapping configuration"
            )
    source_value_datetime = source_value_datetime.astimezone(pytz.utc)
    if config.source.is_gastag_aware and config.is_gas:
        berlin_local_datetime = source_value_datetime.astimezone(_berlin)
        if berlin_local_datetime.time() == dt_module.time(6, 0, 0):
            berlin_local_datetime = berlin_local_datetime.replace(hour=0).replace(tzinfo=None)
            # We need to re-localize the datetime, because the UTC offset might have changed
            # The Gastag does not always start 6h after midnight.
            # It might also be 5h or 7h on DST transition days.
            berlin_local_datetime = _berlin.localize(berlin_local_datetime)
            source_value_datetime = berlin_local_datetime.astimezone(pytz.utc)
    return source_value_datetime


def _convert_aware_datetime_to_target(value: datetime, config: MappingConfig) -> datetime:
    """
    returns a date or datetime object which is compatible with the target system
    """
    if value.tzinfo is None:
        raise ValueError("value must be timezone-aware at this point")
    target_value: datetime = value
    if config.target.is_gastag_aware and config.is_gas:
        _berlin_local_datetime = value.astimezone(_berlin)
        if _berlin_local_datetime.time() == dt_module.time(0, 0, 0):
            _berlin_local_datetime = _berlin_local_datetime.replace(hour=6).replace(tzinfo=None)
            # We need to re-localize the datetime, because the UTC offset might have changed.
            # The Gastag does not always start 6h after midnight.
            # It might also be 5h or 7h on DST transition days.
            _berlin_local_datetime = _berlin.localize(_berlin_local_datetime)
            target_value = _berlin_local_datetime.astimezone(pytz.utc)
    if config.is_end and config.target.is_inclusive_end:
        assert config.target.resolution is not None  # ensured by the consistency check
        target_value = target_value - config.target.resolution  # converts the exclusive end to an inclusive end
        # and e.g. 2024-01-02 00:00:00 to 2024-01-01 23:59:59 if the resolution is timedelta(seconds=1)
        # Work because the original value is - if it is an end - always an exclusive end.
    if config.target.implicit_timezone is not None:
        target_value = target_value.astimezone(config.target.implicit_timezone)
    if config.target.is_date_only:
        target_value = datetime.combine(target_value.date(), datetime.min.time())
    return target_value


def adapt_to_target(source_value: Union[date, datetime], config: MappingConfig) -> datetime:
    """
    maps the source value to a value compatible with the target system by using the given mapping configuration
    """
    if source_value is None:
        raise ValueError("source_value must not be None")
    if config is None:
        raise ValueError("config must not be None")
    if not config.is_self_consistent():
        raise ValueError("config is not self-consistent: " + ", ".join(config.get_consistency_errors()))
    # there are just 2 steps:
    # 1. convert the source from whatever it is to something unified with what we can work
    # 2. convert the unified source to the target (which might be just as obscure as the source)
    source_value_datetime = _convert_source_date_or_datetime_to_aware_datetime(source_value, config)  # step 1
    assert source_value_datetime.tzinfo is not None
    assert source_value_datetime.tzinfo == pytz.utc
    target_value = _convert_aware_datetime_to_target(source_value_datetime, config)  # step 2
    return target_value



"""
Chronomeleon is a Python package that helps you to migrate datetimes from one system to another.
"""

__all__ = ["ChronoAssumption", "MappingConfig", "adapt_to_target"]

from .mapping import adapt_to_target
from .models import ChronoAssumption, MappingConfig



"""
models are the classes used by chronomeleon
"""

__all__ = ["ChronoAssumption", "MappingConfig"]
# fixes: Module "chronomeleon.models" does not explicitly export attribute "ChronoAssumption"

from .chrono_assumption import ChronoAssumption
from .mapping_config import MappingConfig



"""contains the Mapping configuration class"""

from dataclasses import dataclass
from typing import Optional

from .chrono_assumption import ChronoAssumption


@dataclass(frozen=True, kw_only=True)
class MappingConfig:
    """
    represents the mapping rules for one date(time) field from one system to another
    """

    source: ChronoAssumption
    """
    assumptions about the interpretation of the date(time) field in the source system
    """
    target: ChronoAssumption
    """
    assumptions about the interpretation of the date(time) field in the source system
    """

    is_end: Optional[bool] = None
    """
    True if and only if the date or time is the end of a range. None if it doesn't matter.
    """

    is_gas: Optional[bool] = None
    """
    True if the sparte is Gas.
    Set to true to trigger the gas tag modifications in source, target or both, if necessary. Ignore otherwise.
    """

    def get_consistency_errors(self) -> list[str]:
        """
        returns a list of error messages if the mapping configuration is not self-consistent
        """
        errors: list[str] = []
        if not isinstance(self.source, ChronoAssumption):
            errors.append("source must be a ChronoAssumption object")
        else:
            errors.extend(["source: " + x for x in self.source.get_consistency_errors()])
        if not isinstance(self.target, ChronoAssumption):
            errors.append("target must be a ChronoAssumption object")
        else:
            errors.extend(["target: " + x for x in self.target.get_consistency_errors()])
        if (self.source.is_gastag_aware or self.target.is_gastag_aware) and self.is_gas is None:
            errors.append("if is_gastag_aware is set in either source or target, then is_gas must not be None")
            # The opposite is not the case: I can set is_gas to True without setting is_gastag_aware to True
        if (
            self.source.is_inclusive_end is not None or self.target.is_inclusive_end is not None
        ) and self.is_end is None:
            errors.append("if is_inclusive_end is set in either source or target, then is_end must not be None")
        if self.is_end is True and (self.source.is_inclusive_end is None or self.target.is_inclusive_end is None):
            errors.append("if is_end is True, then is_inclusive_end must not be None in both source and target")
        return errors

    def is_self_consistent(self) -> bool:
        """
        checks if the mapping configuration is self-consistent
        """
        return not any(self.get_consistency_errors())



"""contains the ChronoAssumption class"""

from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

from pytz import BaseTzInfo


@dataclass(frozen=True, kw_only=True)
class ChronoAssumption:
    """
    represents assumptions about how a specific system interprets a specific field that holds date or time
    """

    resolution: Optional[timedelta] = None
    """
    This is only necessary to provide, if the field is an inclusive end date.
    The smallest unit of time that this field can represent.
    Typically this is something like 1 day, 1 second, 1 microsecond.
    Adding one "unit" of the resolution leads to the smallest possible increase in the field.
    If e.g. the resolution is 1 day, then the next possible value after 2024-01-01 is 2024-01-02.
    But if the resolution is 1 second, then the next possible value after 2024-01-01 00:00:00 is 2024-01-01 00:00:01.
    """

    implicit_timezone: Optional[BaseTzInfo] = None
    """
    Systems often don't provide an explicit UTC offset with their date or time fields.
    In this case, the system implicitly uses a specific timezone.
    You can specific this implicit timezone here.
    If the datetimes come with a specified UTC offset, leave it None.
    You have to specify the implicit timezone as a pytz-timezone object, e.g. pytz.timezone("Europe/Berlin").
    pytz is a dependency of chronomeleon; If you install chronomeleon, you also get pytz.
    """

    is_inclusive_end: Optional[bool] = None
    """
    Must not be None if is_end is True.
    True if and only if the end of the range is inclusive.
    If the resolution is timedelta(days=1) and is_inclusive_end is True, then the range 2024-01-01 to 2024-01-31 covers
    the entire month of January.
    If is_inclusive_end is False, then the range 2024-01-01 to 2024-02-01 covers the entire month of January.
    """

    is_gastag_aware: bool = False
    """
    True if and only if the start of a day is 6:00 am German local time.
    If you never heard of the "Gastag", you can ignore this parameter and let it default to False.
    """

    is_date_only: bool = False
    """
    True if and only if the field in the respective system is a date without a time component (datetime.date).
    """

    def get_consistency_errors(self) -> list[str]:
        """
        returns errors from the self-consistency check; if the returned list is empty, the object is self-consistent
        """
        result: list[str] = []
        if self.is_inclusive_end and self.resolution is None:
            result.append("if is_inclusive_end is True, then resolution must be set")
        if self.resolution is not None and not isinstance(self.resolution, timedelta):
            result.append(f"resolution must be a timedelta object but is {self.resolution.__class__.__name__}")
        if self.implicit_timezone is not None and not isinstance(self.implicit_timezone, BaseTzInfo):
            result.append(
                f"implicit_timezone must be a pytz timezone object but is {self.implicit_timezone.__class__.__name__}"
            )
        return result

    def is_self_consistent(self) -> bool:
        """
        returns True if the object is self-consistent
        """
        return not any(self.get_consistency_errors())
